/*								TI-89 & 92P Contra!!!
								   by James Dunne
								   
											jdunne4@bradley.edu
											Jaymz031602@netscape.net
	
		As having been one of the greatest games to hit the NES console,
	I decided that a TI-89 port of Contra would be a nice tribute.  This
	source release is the beginning of a Contra game engine.  Further
	development is planned for this game of course, but I decided to release
	the source code now, so as to let others have a shot at making their own
	game before I corrupt it too much with my own game data. ;)
	
		Objects behave like the entity-driven engine of Quake, e.g. Quake-C.
	And entity is everything moving, living, and/or thinking: a player, a bullet,
	an enemy, a trigger, etc.  All entities have the same properties available.

		To spawn an entity, simply create a variable of type (entity_t *) and
	assign it to what the spawn() function returns, like so:  (see NOTE #1)
	
		entity_t  *monster = spawn();
		
	To kill the entity, simply call the kill() function, like so:
		
		kill(monster);
		
	Easy enough!
		
		See the definition of entity_t for different properties of entities.  Of course,
	there are coordinates {x, y} and there are velocities {dx, dy}. (see NOTE #2!! IMPORTANT)
	There is a lot more stuff here than I care to explain, but if you're daring enough, you
	can figure it out eventually ;).  Most of the code is pure C, and very straight-forward.


NOTE #1: You should only spawn entities when they come on the screen. I'll leave
	it up to you to figure out how to do that cleverly ;).  A maximum of 24 entities
	is allowed on the screen at one time.  This can be changed of course.  Simply
	change the "#define MAX_ENTITIES	24" to something else.  The array of entities
	is allocated dynamically at the beginning of _main().  This must be done before
	any spawn(), kill(), etc functions are used.
		
	
NOTE #2: Coordinates {x, y} are stored shifted by 8 bits to the left in order
	to maintain a little bit of realism, so 1 pixel on the TI = 256 units in my
	coordinate system.  This allows for pseudo-decimal increments in position
	without having to use the incredibly slow floating point routines on the TI.
*/


#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define SAVE_SCREEN           // Save/Restore LCD Contents
#include <tigcclib.h>         // Include All Header Files

#define bool	unsigned char

// Compile for TI-92P
//short _ti92plus;              // Produce .9XZ File
//#define __MYTI92P
//#define __LINES__		128

// Compile for TI-89
// this is the only calculator I own, so :-P
short _ti89;                  // Produce .89Z File
#define __MYTI89
#define	__LINES__		100


// Use my own defines for the keyboard buttons:
//  Keeps up portability between TI-89 and 92.
#ifdef __MYTI89
	#define	KEYB_UP		(1<<0)
	#define KEYB_LEFT	(1<<1)
	#define KEYB_DOWN	(1<<2)
	#define KEYB_RIGHT	(1<<3)
	
	#define KEYB_2ND	(1<<4)
	#define KEYB_SHIFT	(1<<5)
	#define KEYB_DIA	(1<<6)
	#define KEYB_ALPHA	(1<<7)
	
	#define KEYB_F1_ROW		5
	#define KEYB_F1			(1<<7)
	#define KEYB_F2_ROW		4
	#define KEYB_F2			(1<<7)
	
	#define KEYB_ESC_ROW	6
	#define KEYB_ESC_MASK	(1<<0)
	#define KEYB_ENTER_ROW	1
	#define KEYB_ENTER		(1<<0)

	#define KEYB_GRAY_ROW		1
	#define KEYB_GRAY_BITUP		(1<<1)
	#define KEYB_GRAY_BITDOWN	(1<<2)

	#define	TILE_W		20
	#define	TILE_H		12
	// Status line start:
	#define	LASTLINE	96

	#define	screen_w	160
	#define screen_h	__LINES__
#endif

// For the TI-92P keyboard:
#ifdef __MYTI92P
	#define KEYB_LEFT	(1<<4)
	#define	KEYB_UP		(1<<5)
	#define KEYB_RIGHT	(1<<6)
	#define KEYB_DOWN	(1<<7)
	
	#define KEYB_2ND	(1<<0)
	#define KEYB_DIA	(1<<1)
	#define KEYB_SHIFT	(1<<2)
	// This is actually the HAND key:
	#define KEYB_ALPHA	(1<<3)
	
	#define KEYB_F1_ROW		6
	#define KEYB_F1			(1<<4)
	#define KEYB_F2_ROW		4
	#define KEYB_F2			(1<<4)

	#define KEYB_ESC_ROW	8
	#define KEYB_ESC_MASK	(1<<6)
	#define KEYB_ENTER_ROW	9
	#define KEYB_ENTER		(1<<1)

	// FIXME:  Change these:
	#define KEYB_GRAY_ROW		1
	#define KEYB_GRAY_BITUP		(1<<1)
	#define KEYB_GRAY_BITDOWN	(1<<2)

	#define	TILE_W		30
	#define	TILE_H		16
	// Status line start:
	#define LASTLINE	128

	#define	screen_w	240
	#define screen_h	__LINES__
#endif

unsigned char *backbufl = NULL, *backbufd = NULL;
unsigned char *sprtbufl = NULL, *sprtbufd = NULL;
unsigned char *light_plane = NULL, *dark_plane = NULL;

bool	link_play = FALSE;
bool	is_client = FALSE, is_server = TRUE;

#include "linkplay.h"
#include "rle.c"

#include "sprites.c"
#ifdef __MYTI89
#include "titlescr.h"
#else
#include "title92scr.h"
#endif
#include "selectorspr.h"

// Although this looks like a horrible hack, its actually better to do this
// with TIGCC.  It works nicely for me, and is divided into logical subdivisions
// NOTE:  The order of #include's DOES matter.
#include "scroll.c"
#include "entities.c"
#include "bullet.c"
#include "player.c"
#include "grunt.c"

// Returns TRUE if using 2-player link-mode, FALSE for single-player.
bool main_menu() {
	int done = FALSE;
	bool twoplayer = FALSE;
	bool scrolling = TRUE;
	int	scrollx = 0;
	int y = 0, i = 0, p = 0;
	
	scrolling = TRUE;
	scrollx = 0;
	
	GrayAdjust(0);
	
	memset(light_plane, 0xFF, 3840);
	memset(dark_plane, 0xFF, 3840);

	while (!done) {
		// Adjust the grayscale:
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITUP)
			GrayAdjust(grayadjust--);
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITDOWN)
			GrayAdjust(grayadjust++);

		// F1, UP = 1 player
		if (((_rowread(~(1<<KEYB_F1_ROW)) & KEYB_F1) == KEYB_F1) || 
		    ((_rowread(~(1<<0)) & KEYB_UP) == KEYB_UP))
		{
			scrollx = screen_w - 1;
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			
			Sprite16(4, 78, 10, selector_l, light_plane, SPRT_OR);
			Sprite16(4, 78, 10, selector_d, dark_plane, SPRT_OR);
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,100}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,100}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);

			twoplayer = FALSE;
		};
		
		// F2, DOWN = 2 player
		if (((_rowread(~(1<<KEYB_F2_ROW)) & KEYB_F2) == KEYB_F2) || 
		    ((_rowread(~(1<<0)) & KEYB_DOWN) == KEYB_DOWN))
		{
			scrollx = screen_w - 1;
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,99}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,90,19,99}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
			
			Sprite16(4, 90, 10, selector_l, light_plane, SPRT_OR);
			Sprite16(4, 90, 10, selector_d, dark_plane, SPRT_OR);
			
			PortSet(light_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);
			PortSet(dark_plane, 239, 139);
			ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_NORMAL);

			twoplayer = TRUE;
		};

		if (scrolling && (y++ >= 50)) {
			// Scroll the title screen in from the right:
			ScrollLeft((unsigned short *)light_plane, __LINES__);
			ScrollLeft((unsigned short *)dark_plane, __LINES__);
			// Add a new line on the right side of the screen:
		#ifdef __MYTI89
			p = 19;
		#else
			p = 29;
		#endif
			i = (scrollx >> 3);
			for (y=0;y<__LINES__;y++) {
				((unsigned char *)light_plane)[p] &= ~1;
				((unsigned char *)dark_plane)[p] &= ~1;
				((unsigned char *)light_plane)[p] |= (titlescr_l[i] & (1 << (7 - (scrollx & 7)))) >> (7 - (scrollx & 7));
				((unsigned char *)dark_plane)[p] |= (titlescr_d[i] & (1 << (7 - (scrollx & 7)))) >> (7 - (scrollx & 7));
				p += 30;
				i += 30;
			};
			if (scrollx++ >= screen_w) {
				scrolling = FALSE;
				memcpy(light_plane, titlescr_l, 3840);
				memcpy(dark_plane, titlescr_d, 3840);

				// Draw one-player selector
				PortSet(light_plane, 239, 139);
				ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
				PortSet(dark_plane, 239, 139);
				ScrRectFill(&(SCR_RECT){{4,78,19,87}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);
				
				Sprite16(4, 78, 10, selector_l, light_plane, SPRT_OR);
				Sprite16(4, 78, 10, selector_d, dark_plane, SPRT_OR);
			}
			y = 0;
		};

		// Frame rate limiter:
		if (!scrolling) {
			OSFreeTimer(6);
			OSRegisterTimer(6, 1);
		};

		if (!scrolling) while (!OSTimerExpired(6)) ;

		// 2ND or ENTER accepts choice:
		if ((_rowread(~(1<<0)) & KEYB_2ND) == KEYB_2ND) break;
		if ((_rowread(~(1<<KEYB_ENTER_ROW)) & KEYB_ENTER) == KEYB_ENTER) break;
		
		// ESC quits at any time:
		if ((_rowread(~(1<<KEYB_ESC_ROW)) & KEYB_ESC_MASK) == KEYB_ESC_MASK) break;
	};
	
	return twoplayer;
};

void map_save(const char name[]) {
	const unsigned char map_file_extension[6] = { 0,'C','M','P',0,0xf8 };
	FILE *fnew = fopen(name+1, "wb");
	if (fnew != NULL) {
		// Write out compressed map data:
		frleputc( 0, fnew, 2 );
		frlewrite(map, TILE_W, TILE_H, fnew);
		frleputc( 0, fnew, 1 );
		// Write file extension:
		fwrite(map_file_extension, 6, 1, fnew);
		fclose(fnew);
		// Archive the file:
		EM_moveSymToExtMem( name+9, HS_NULL );
	};
};

void map_load(const char name[]) {
	FILE *fnew = fopen(name, "rb");
	if (fnew != NULL) {
		frleread(map, TILE_W, TILE_H, fnew);
		fclose(fnew);
	};
};

// Main Function
void _main(void) {
	int 	i, j;
	long	oldscrn_x = 0, oldscrn_y = 0;

	time = 0;
	horiz_scroll = TRUE;
	
	// Initialize all memory before going into "graphics mode":
	
	// Maybe in a future release, use the gray.h GrayDBuf...() functions instead of
	// this old way:  (might save some memory in the process as well??)
	backbufl = (unsigned char *)safe_malloc(LCD_SIZE);
	if (backbufl == NULL) return;
	backbufd = (unsigned char *)safe_malloc(LCD_SIZE);
	if (backbufd == NULL) return;
	sprtbufl = (unsigned char *)safe_malloc(LCD_SIZE);
	if (sprtbufl == NULL) return;
	sprtbufd = (unsigned char *)safe_malloc(LCD_SIZE);
	if (sprtbufd == NULL) return;
	
	ClrScr();

	INT_HANDLER save_int_1 = GetIntVec (AUTO_INT_1);
	SetIntVec (AUTO_INT_1, DUMMY_HANDLER);   // redirect auto-int 1 to "nothing"
	// enable grayscale
	GrayOn();									//turn grayscale on

	ScrRectFill(&(SCR_RECT){{0,0,160,100}}, &(SCR_RECT){{0,0,160,100}}, A_REVERSE);		//turns screen white
	
	light_plane = GetPlane(LIGHT_PLANE);
	dark_plane = GetPlane(DARK_PLANE);
	
	// Main-menu: (returns TRUE if using two-player cable-link mode)
	link_play = main_menu();

	GrayAdjust(grayadjust = 0);

	// Initialization of game-data:
	entity_t	*curr;

	// Must be called before any spawn(), kill(), etc...
	entities = (entity_t **)calloc(MAX_ENTITIES, sizeof(entity_t *));
	
	map	= (unsigned char *)calloc(TILE_W * TILE_H, sizeof(unsigned char));
	if (map == NULL) return;
	unsigned char *mapi = map;

	player = player_spawn();
	player->oldx = player->x = 2048;
	player->oldy = player->y = 2048;
	unsigned char playerindex = player->index;

	// Messages and other immediate drawing commands are re-routed to the dark-sprite buffer
	// and will only be displayed when the sprite buffer is swapped to the screen
	PortSet (sprtbufd, 239, 127);

	// Draw initial background:
	memset(backbufl, 0b00000000, LCD_SIZE);
	memset(backbufd, 0b00000000, LCD_SIZE);

	randomize();

	// Background screen-scrolling coordinates defining the upper-left corner of
	// the view screen.
	screen_x = 0;
	screen_y = 0;
	
	// Draw first background layout:
	int x, y, s = 0;
	
	// The global (unsigned char *)map variable contains a TILE_W x TILE_H tile-map
	// of the CURRENT VISIBLE SCREEN.  Each tile # represents the corresponding
	// 8x8 tile index in the backsprites[] array defined in 'sprites.c'
	mapi = map;
	for (x=0; x<TILE_W; x++) {
		for (y=0; y<TILE_H; y++) {
			// Simple one-screen test map used to create & debug collision detection and scrolling.
			if ((y == 0) || (y == TILE_H-1)) *mapi = 1 | 0x80; else *mapi = 0;
			if (((x % 4) == 0) && (y == (TILE_H - 2))) *mapi = 1 | 0x80;
			s = *mapi++ & 0x7f;
			Sprite8(x << 3, y << 3, 8, backsprites[(s << 1) + 0], backbufl, A_OR);
			Sprite8(x << 3, y << 3, 8, backsprites[(s << 1) + 1], backbufd, A_OR);
		};
	};

	// Works, but is a bit kludgy sometimes:
	/*
	// Write out map:
	map_save("\0CONTRA  ");
	
	// Clear out map:
	memset(map, 0, TILE_W * TILE_H);
	
	map_load("CONTRA  ");
	*/
	packetheader_t	send_hdr, recv_hdr;

	// Now, determine who is server and who is client:
	if (link_play) {
		is_client = FALSE;
		is_server = FALSE;
		for (i=0;i<10;i++) {
			// Request a game packet:
			send_hdr.type = PACKET_GAME_REQUEST;
			LIO_SendData(&send_hdr, sizeof(packetheader_t));
			// Wait 1 second to receive data:
			if (LIO_RecvData(&recv_hdr, sizeof(packetheader_t), 40) == 0) {
				// Got a packet header!
				// Got the entire packet:
				if (recv_hdr.type == PACKET_GAME_REQUEST) {
					send_hdr.type = PACKET_GAME_SERVER;
					LIO_SendData(&send_hdr, sizeof(packetheader_t));
					is_client = FALSE;
					is_server = TRUE;
					break;
				} else if (recv_hdr.type == PACKET_GAME_SERVER) {
					is_client = TRUE;
					is_server = FALSE;
					break;
				};
			};
		};
		// I am just a simple client who sends keystrokes to the server and retrieves
		// necessary entity information:
		if (is_client) {
			serverplayer = player_spawn();
			serverplayer->think = NULL;
			player->think = (func_t *)player_client_think;
		}
		if (is_server) {
			entity_t *clientplayer = player_spawn();
			clientplayer->think = (func_t *)player_server_think;
		}
		if (!is_server && !is_client) {
			link_play = FALSE;
		};
	};


	// TAKE NOTE:
		// Actual map data should be stored in column-down, row format to easily read in the new
		// line to scroll (horizontal scrolling only). Vertical scrolling maps will be stored the
		// opposite way.
	
	oldscrn_x = screen_x;
	oldscrn_y = screen_y;
	
	// Main loop: (until player is dead)
	while (entities[playerindex] != NULL) {
		// Frame rate limiter:
		OSFreeTimer(6);
		OSRegisterTimer(6, 1);

		// Adjust the grayscale (+ and -):
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITUP)
			GrayAdjust(grayadjust--);
		if (_rowread(~(1<<KEYB_GRAY_ROW)) & KEYB_GRAY_BITDOWN)
			GrayAdjust(grayadjust++);
		
		// Maybe implement switchable scrolling methods i.e. original Contra
		// one level is vertical scrolling only, another is horizontal scrolling only.
		// vertical scrolling can be up or down, horizontal only right
		
		// Default is side-scrolling, in fact the map handling code will
		// be biased towards loading the rightmost row in order to scroll
		// the map correctly:
		
		// Maybe we should implement a left-scrolling algorithm as well for some
		// levels, so we don't get the original Contra's feeling of heading EAST all
		// the time.
		
		if (horiz_scroll) {
			// Scroll the background left one line:
			if (screen_x > oldscrn_x) {
				for (i=oldscrn_x; i<screen_x; i++) {
					ScrollLeft((unsigned short *)backbufl, __LINES__);
					ScrollLeft((unsigned short *)backbufd, __LINES__);
				};
				// Determing starting row to draw from:
				// Using two pointers for this is assumedly faster access due
				// to not having to calculate the index for each column, since
				// we already know the width of the visible map:
				mapi = &( map[ ((screen_x >> 3) % TILE_W) * TILE_H] );
				for (y=0; y<TILE_H; y++) {
					s = *mapi & 0x7F;
					Sprite8((TILE_W << 3) - (screen_x % 8), (y << 3), 8, backsprites[(s << 1) + 0], backbufl, A_OR);
					Sprite8((TILE_W << 3) - (screen_x % 8), (y << 3), 8, backsprites[(s << 1) + 1], backbufd, A_OR);
					// Next column:
					mapi++;
				};
			}
		} /*else {
			// Not tested, but theoretically should be correct:
			// Note, if you want up AND downward scrolling levels,
			// you'll have to have a global control flag which
			// specifies which direction to scroll and change the
			// code below accordingly:
			if (screen_x > oldscrn_x) {
				for (i=oldscrn_y; i<screen_y; i++) {
					ScrollDown((unsigned short *)backbufl, __LINES__);
					ScrollDown((unsigned short *)backbufd, __LINES__);
				};
				mapi = map[TILE_W - 1];
				for (x=0; x<TILE_H; y++) {
					s = *mapi;
					Sprite8((x << 3) - (screen_x % 8), 0 - (screen_y % 8), 8, backsprites[(s << 1) + 0], backbufl, A_OR);
					Sprite8((x << 3) - (screen_x % 8), 0 - (screen_y % 8), 8, backsprites[(s << 1) + 1], backbufd, A_OR);
					mapi += TILE_W;
				};
			}
		}*/
		
		// The background is copied into the sprite-buffer:
		memcpy(sprtbufl, backbufl, LCD_SIZE);
		memcpy(sprtbufd, backbufd, LCD_SIZE);

		// Prevent unsigned int from overflowing
		if (player->x < screen_x + 256) player->x = screen_x + 256;
		if (player->y < screen_y + 256) player->y = screen_y + 256;

		oldscrn_x = screen_x;
		oldscrn_y = screen_y;

		if ((link_play && is_server) || (!link_play)) {
			if (time % 120 == rand() % 70) {
				spawn_grunt();
			};
		};
		
		// Handle thinking, collisions, and rendering for every entity:
		for (i = 0; i < MAX_ENTITIES; i++) if (entities[i] != NULL) {
			int tx, ty;
			curr = entities[i];
			
			if ((link_play && is_server) || (!link_play)) {
				// Check for collisions among entities:
				for (j = 0; j < MAX_ENTITIES; j++) if ((entities[j] != NULL) && (j != i)) {
					if (entities[i] == NULL) break;
					if (collide(entities[i], entities[j])) {
						if (entities[i]->touch) entities[i]->touch(entities[i], entities[j]);
						if (entities[i] == NULL) break;
						if (entities[j]->touch) entities[j]->touch(entities[j], entities[i]);
					}
				};
			};
			
			if (entities[i] == NULL) continue;
			
			// Think for the entity (if it's time):
			if ((link_play && is_server) || (!link_play))
				if ((curr->think != NULL) && (time == curr->next_think))
					curr->think(curr);
			
			// We have to check if this still exists, because the entity
			// could have removed itself via kill(self):
			// We can NOT check if curr == NULL because curr (as a pointer) is not passed
			// by reference, and therefore will not reflect the changes to its master
			// pointer in entites[], unless we use curr as a double-pointer?? :-D
			if (entities[i] == NULL) continue;

			tx = curr->x >> 8;
			ty = curr->y >> 8;

			// Check if we should bother to draw first:
			if (curr->render != NULL)
				// Is it within screen bounds?
				if ((tx >= screen_x) && (tx <= screen_x + screen_w) && 
					(ty >= screen_y) && (ty <= screen_y + screen_h)) 
					curr->render(curr);
				
			if ((link_play && is_server) || (!link_play)) {
				// Apply the deltas: (velocity)
				curr->oldx = curr->x;
				curr->oldy = curr->y;
				curr->x += curr->dx;
				curr->y += curr->dy;
				
				// Check for ground & wall collisions:
				if (curr->collide != NULL) {
					curr->colflags = check_map(curr);
					curr->collide(curr);
				};
				
				if (entities[i] == NULL) continue;
	
				tx = curr->x >> 8;
				ty = curr->y >> 8;
	
				// Off the edge of the screen?  Kill it!
				if (curr->class != CLASS_PLAYER) {
					if ((tx < screen_x - curr->width - 4) ||
						(tx > 4 + screen_x + screen_w) ||
						(ty < screen_y - curr->height - 4) ||
						(ty > 4 + screen_y + screen_h))
					{
						kill(curr);
						continue;
					}
				} else {
					if (tx < screen_x) curr->x = screen_x << 8;
					if (tx > screen_x + screen_w - curr->width)
						curr->x = (screen_x + screen_w - curr->width) << 8;
				};
				
				if (entities[i] == NULL) continue;
				
				// Handle scrolling the screen for CLASS_PLAYER's:
				if (curr->class == CLASS_PLAYER) {
					if (tx - screen_x > 80) screen_x = (tx - 80);
					if (screen_x < 0) screen_x = 0;
				}
				
				// Gravity (only when not ONGROUND and not NOGRAVITY):
				if (((curr->flags & FLAG_ONGROUND) != FLAG_ONGROUND) && ((curr->flags & FLAG_NOGRAVITY) != FLAG_NOGRAVITY))
					curr->dy += 100;
				
				// Terminal velocity:
				if (curr->dy >= 2048) curr->dy = 2048;
				
				// Friction (only when not NOFRICTION):
				if ((curr->flags & FLAG_NOFRICTION) != FLAG_NOFRICTION) {
					// Since we have scaled everything up by 256, we can't multiply
					// by 0.1 or anything small like that, because they're not decimals,
					// so, alas, divides will have to suffice.  But wait!  Right-Shifts
					// work FASTER!!!
					if (curr->dx < -32) curr->dx = -(abs(curr->dx) >> 1);
					else if (curr->dx > 32) curr->dx >>= 1;
					else curr->dx = 0;
				};

				if (link_play) {
					// Handle special case for bullets, etc...
					// Implement fire-and-forget code
					send_hdr.type = PACKET_ENTITY_SERVER_UPDATE;
					send_hdr.length = sizeof(netentity_t);
					LIO_SendData(&send_hdr, sizeof(packetheader_t));
					LIO_SendData(&(curr->index), sizeof(int));
					LIO_SendData((netentity_t *)curr, sizeof(netentity_t));
				};
			};
		};
		if (link_play && is_server) {
			// Send the end-of-frame packet:
			send_hdr.type = PACKET_END_FRAME;
			send_hdr.length = 0;
			LIO_SendData(&send_hdr, sizeof(packetheader_t));
			
			// Wait for the packet update from the client:
			if (LIO_RecvData(&recv_hdr, sizeof(packetheader_t), 0)) {
				if (recv_hdr.type == PACKET_CLIENT_UPDATE) {
					LIO_RecvData(&(clientplayer->buttons), recv_hdr.length, 0);
				};
			};
		};
		if (link_play && is_client) {
			// Client update code: (simply control the player and update buttons to the server):
			send_hdr.type = PACKET_CLIENT_UPDATE;
			send_hdr.length = sizeof(unsigned char);
			LIO_SendData(&send_hdr, sizeof(packetheader_t));
			LIO_SendData(&(player->buttons), sizeof(unsigned char));
			
			// Retrieve entity info:
			while (recv_hdr.type != PACKET_END_FRAME) {
				if (LIO_RecvData(&recv_hdr, sizeof(packetheader_t), 0) == 0) {
					if (recv_hdr.type == PACKET_ENTITY_SERVER_UPDATE) {
						int index;
						LIO_RecvData(&index, sizeof(int), 1);
						LIO_RecvData((netentity_t *)entities[index], sizeof(netentity_t), 1);
					} else if (recv_hdr.type == PACKET_ENTITY_SPAWN) {
						int index;
						LIO_RecvData(&index, sizeof(int), 1);
						entities[index] = (entity_t *)malloc(sizeof(entity_t));
						entities[index]->index = index;
					};
				};
			};
		};

		// The sprites & background are rendered in sprtbufl & d.
		// Copy them to the LCD
		memcpy(light_plane, sprtbufl, LCD_SIZE);
		memcpy(dark_plane, sprtbufd, LCD_SIZE);

		// Frame rate limiter:
		while (!OSTimerExpired(6)) ;
		
		time++;
		
		// ESC quits at any time:
		if ((_rowread(~(1<<KEYB_ESC_ROW)) & KEYB_ESC_MASK) == KEYB_ESC_MASK) break;
	};
	
	// Remove all entities:
	for (i = 0; i < MAX_ENTITIES; i++) {
		kill(entities[i]);
		entities[i] = NULL;
	};
	
	free(map);

	free(entities);
	
	free(sprtbufl);
	free(sprtbufd);

	free(backbufl);
	free(backbufd);
	
	GrayAdjust(0);
	
	GrayOff();
	
	// disable grayscale
	SetIntVec (AUTO_INT_1, save_int_1);
	
	ClrScr();
	return;
};
