// Player handling routines:
void draw_player(struct entity_t *self) {
	// This is the general function to simply render a player-looking
	// sprite:
	if (self == NULL) return;
	int x = (self->x >> 8) - screen_x;
	int y = (self->y >> 8) - screen_y;

	if (((self->flags & FLAG_LIEDOWN) && !(self->buttons & KEYB_LEFT) && !(self->buttons & KEYB_RIGHT) &&
	     (self->flags & FLAG_ONGROUND)) || (self->flags & FLAG_DEAD))
	{
		if (self->flags & FLAG_FACERIGHT) {
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 1], sprtbufd, A_OR);
			Sprite8(x + 8, y + 8, 8, sprites_player[(0x02 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x + 8, y + 8, 8, sprites_player[(0x02 << 2) + 1], sprtbufd, A_OR);
		} else {
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x03 << 2) + 3], sprtbufd, A_OR);
			Sprite8(x - 8, y + 8, 8, sprites_player[(0x02 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x - 8, y + 8, 8, sprites_player[(0x02 << 2) + 3], sprtbufd, A_OR);
		};
	} else {
		if (self->flags & FLAG_FACERIGHT) {
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 1], sprtbufd, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 0], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 1], sprtbufd, A_OR);
		} else {
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x, y, 8, sprites_player[(0x00 << 2) + 3], sprtbufd, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 2], sprtbufl, A_OR);
			Sprite8(x, y + 8, 8, sprites_player[(0x01 << 2) + 3], sprtbufd, A_OR);
		};		
	};
};

// Wrapper function called from main-loop to render a player:
void render_player(struct entity_t *self) {
	int i;

	// Render health bar:
	for (i=0;i<(self->health);i++) {
		Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[0], sprtbufl, A_OR);
		Sprite8(8 + (i << 3), 8, 8, sprites_healthbar[1], sprtbufd, A_OR);
		Sprite8(8 + (i << 3), 16, 8, sprites_healthbar[2], sprtbufl, A_OR);
		Sprite8(8 + (i << 3), 16, 8, sprites_healthbar[3], sprtbufd, A_OR);
	};

	// Make player blink when invulnerable:
	if ((self->invuln_time >= time) && ((time % 2) == 0)) return;
	// Draw the player:
	draw_player(self);
};

// Called to fire the player's weapon:
void player_fire_weapon(entity_t *self) {
	switch (self->weapon) {
		case 0: 	// Regular shotgun
			// No repeated fire:
			if (self->lastbuttons & KEYB_2ND) break;
			self->attack_time = time + 4;
			// Determine which state we are in so that we may fire in
			// the appropriate direction
			if (self->flags & FLAG_LOOKUP) {
				signed long dx = 0, dy = -880, yo = -2048, xo = 0;
				if (self->buttons & KEYB_LEFT) { dx = -622; dy = -622; xo = -2048; yo = 0; }
				if (self->buttons & KEYB_RIGHT) { dx = 622; dy = -622; xo = 2048; yo = 0; }
				spawn_bullet(self, xo, yo, dx, dy);
			} else if (self->flags & FLAG_LIEDOWN) {
				signed long dx = 0, dy = 880, yo = 2048, xo = 0;
				if (self->buttons & KEYB_LEFT) { dx = -622; dy = 622; xo = -2048; }
				else if (self->buttons & KEYB_RIGHT) { dx = 622; dy = 622; xo = 2048; }
				else if (!(self->flags & FLAG_ONGROUND)) {
					xo = 0; yo = (16 << 8); dy = 880; dx = 0;
				} else {
					xo = (self->flags & FLAG_FACERIGHT) ? 2048 : -2048;
					dx = (self->flags & FLAG_FACERIGHT) ? 1280 : -1280;
					dy = 0;
				};
				spawn_bullet(self, xo, yo, dx, dy);
			} else {
				spawn_bullet(self, (self->flags & FLAG_FACERIGHT) ? 2048 : -2048, 8 << 8,
							 ((self->flags & FLAG_FACERIGHT) ? 880 : -880), 0);
			};
			break;
		case 1: 	// Machinegun
			// Repeated fire allowed (the purpose of a machine-gun)
			self->attack_time = time + 3;
			// Determine which state we are in so that we may fire in
			// the appropriate direction
			if (self->flags & FLAG_LOOKUP) {
				signed long dx = 0, dy = -1280, yo = -2048, xo = 0;
				if (self->buttons & KEYB_LEFT) { dx = -905; dy = -905; xo = -2048; yo = 0; }
				else if (self->buttons & KEYB_RIGHT) { dx = 905; dy = -905; xo = 2048; yo = 0; }
				spawn_bullet(self, xo-1, yo-1, dx, dy);
			} else if (self->flags & FLAG_LIEDOWN) {
				signed long dx = 0, dy = 1280, yo = 2048, xo = 0;
				if (self->buttons & KEYB_LEFT) { dx = -905; dy = 905; xo = -2048; }
				else if (self->buttons & KEYB_RIGHT) { dx = 905; dy = 905; xo = 2048; }
				else if (!(self->flags & FLAG_ONGROUND)) {
					xo = 0;	yo = (16<<8); dy = 1280; dx = 0;
				} else {
					xo = (self->flags & FLAG_FACERIGHT) ? 2048 : -2048;
					dx = (self->flags & FLAG_FACERIGHT) ? 1280 : -1280;
					dy = 0;
				};
				spawn_bullet(self, xo, yo, dx, dy);
			} else {
				// Nothing fancy, just shoot left or right:
				spawn_bullet(self, ((self->flags & FLAG_FACERIGHT) ? 
								   (((self->flags & FLAG_LIEDOWN) ? 16 : 8) << 8) : 
								  -(((self->flags & FLAG_LIEDOWN) ? 16 : 8) << 8)),
							 8 << 8, ((self->flags & FLAG_FACERIGHT) ? 1280 : -1280), 0);
			};
			break;
		case 2: // Spread
			// No repeated fire:
			if (self->lastbuttons & KEYB_2ND) break;
			
			self->attack_time = time + 8;
			
			// Spread angles at -30 degrees, -15 degrees, 0 degrees, 15 degrees, and 30 degrees
			// for 5 bullets.  Speed is set at 880 units/frame
			
			// Values were calculated with {880 * cos(angle + delta), 880 * sin(angle + delta)}
			// where delta is the set of {-30, -15, 0, 15, 30} and angle is the implied angle of
			// the gun, 0 being right, 90 being upwards, 180 being left, and 270 being downwards.
			
			if (self->flags & FLAG_LOOKUP) {
				if (self->buttons & KEYB_LEFT) {
					// Shoot up-left:		(angle = 135)
					spawn_bullet(self, -2048, 0, -850, -227);
					spawn_bullet(self, -2048, 0, -762, -440);
					spawn_bullet(self, -2048, 0, -622, -622);
					spawn_bullet(self, -2048, 0, -440, -762);
					spawn_bullet(self, -2048, 0, -227, -850);
				} else if (self->buttons & KEYB_RIGHT) {
					// Shoot up-right:		(angle = 45)
					spawn_bullet(self, 2048, 0, 850, -227);
					spawn_bullet(self, 2048, 0, 762, -440);
					spawn_bullet(self, 2048, 0, 622, -622);
					spawn_bullet(self, 2048, 0, 440, -762);
					spawn_bullet(self, 2048, 0, 227, -850);
				} else {
					// Shoot straight up:	(angle = 90)
					spawn_bullet(self, 0, 0, -440, -762);
					spawn_bullet(self, 0, 0, -227, -850);
					spawn_bullet(self, 0, 0, 0, -880);
					spawn_bullet(self, 0, 0, 227, -850);
					spawn_bullet(self, 0, 0, 440, -762);
				};
			} else if (self->flags & FLAG_LIEDOWN) {
				if (self->buttons & KEYB_LEFT) {
					// Shoot down-left:		(angle = 225)
					spawn_bullet(self, -2048, 2048, -850, 227);
					spawn_bullet(self, -2048, 2048, -762, 440);
					spawn_bullet(self, -2048, 2048, -622, 622);
					spawn_bullet(self, -2048, 2048, -440, 762);
					spawn_bullet(self, -2048, 2048, -227, 850);
				} else if (self->buttons & KEYB_RIGHT) {
					// Shoot down-right:	(angle = 315)
					spawn_bullet(self, 2048, 2048, 850, 227);
					spawn_bullet(self, 2048, 2048, 762, 440);
					spawn_bullet(self, 2048, 2048, 622, 622);
					spawn_bullet(self, 2048, 2048, 440, 762);
					spawn_bullet(self, 2048, 2048, 227, 850);
				} else if (!(self->flags & FLAG_ONGROUND)) {
					// Shoot straight down:	(angle = 270)	(only if in air)
					spawn_bullet(self, 0, (16<<8), -440, 762);
					spawn_bullet(self, 0, (16<<8), -227, 850);
					spawn_bullet(self, 0, (16<<8), 0, 880);
					spawn_bullet(self, 0, (16<<8), 227, 850);
					spawn_bullet(self, 0, (16<<8), 440, 762);
				} else {
					// Shoot straight left or right:	(angle = 0 or 180)
					if (self->flags & FLAG_FACERIGHT) {
						spawn_bullet(self, 2048, 2048, 762, -440);
						spawn_bullet(self, 2048, 2048, 850, -227);
						spawn_bullet(self, 2048, 2048, 880, 0);
						spawn_bullet(self, 2048, 2048, 850, 227);
						spawn_bullet(self, 2048, 2048, 762, 440);
					} else {
						spawn_bullet(self, -2048, 2048, -762, -440);
						spawn_bullet(self, -2048, 2048, -850, -227);
						spawn_bullet(self, -2048, 2048, -880, 0);
						spawn_bullet(self, -2048, 2048, -850, 227);
						spawn_bullet(self, -2048, 2048, -762, 440);
					};
				};
			} else {
				// Shoot straight left or right:	(angle = 0 or 180)
				if (self->flags & FLAG_FACERIGHT) {
					spawn_bullet(self, 2048, 2048, 762, -440);
					spawn_bullet(self, 2048, 2048, 850, -227);
					spawn_bullet(self, 2048, 2048, 880, 0);
					spawn_bullet(self, 2048, 2048, 850, 227);
					spawn_bullet(self, 2048, 2048, 762, 440);
				} else {
					spawn_bullet(self, -2048, 2048, -762, -440);
					spawn_bullet(self, -2048, 2048, -850, -227);
					spawn_bullet(self, -2048, 2048, -880, 0);
					spawn_bullet(self, -2048, 2048, -850, 227);
					spawn_bullet(self, -2048, 2048, -762, 440);
				};
			};
			break;
		case 3: // Fire
		case 4: // Laser
			break;
	};
};	

void player_server_think(entity_t *self) {
	// Cycle weapons with SHIFT key:		(TEMPORARY)
	if ((self->buttons & KEYB_SHIFT) && !(self->lastbuttons & KEYB_SHIFT)) {
		self->weapon = (self->weapon + 1) % 3;
	};
	
	// Handle buttons:
	if ((self->buttons & KEYB_DIA) && !(self->lastbuttons & KEYB_DIA)) {
		if (self->flags & FLAG_ONGROUND) {
			self->dy -= 1024; self->flags &= ~FLAG_ONGROUND;
		}
	};
	if ((self->buttons & KEYB_2ND) && (time >= self->attack_time)) {
		player_fire_weapon(self);
	};
	
	if (self->buttons & KEYB_UP) {
		self->flags |= FLAG_LOOKUP;
		self->flags &= ~FLAG_LIEDOWN;
	} else self->flags &= ~FLAG_LOOKUP;
	
	if (self->buttons & KEYB_DOWN) {
		self->flags |= FLAG_LIEDOWN;
	} else {
		self->flags &= ~FLAG_LIEDOWN;
	}

	if (self->buttons & KEYB_LEFT)	{
		self->dx = -420;
		self->flags &= ~FLAG_FACERIGHT;
	};
	if (self->buttons & KEYB_RIGHT)	{
		self->dx = 420;
		self->flags |= FLAG_FACERIGHT;
	};

	self->next_think = time + 1;
	self->lastbuttons = self->buttons;
};
	
// Basic control thinking function for the player:  (handle buttons)
void player_think(struct entity_t *self) {
	signed long dx45, dx22, dy45, dy22;
	if (self == NULL) return;

	// Read control buttons (thankfully the same row on TI-89 and the 92+):
	self->buttons = _rowread(~(1<<0));
	player_server_think(self);
};

void player_client_think(entity_t *self) {
	self->buttons = _rowread(~(1<<0));
	self->next_think = time + 1;
};

// What it's like when players collide (into the map)
void player_collide(struct entity_t *self) {
	if ((self->colflags & BLOCKED_GROUND)) {
		self->y = (self->y >> 11) << 11;
		self->dy = 0;
		self->flags |= FLAG_ONGROUND;
	} else {
		self->flags &= ~FLAG_ONGROUND;
	};
	if ((self->colflags & BLOCKED_LHS) == BLOCKED_LHS) {
		self->x = self->oldx;
		self->dx = 0;
		return;
	};
	if ((self->colflags & BLOCKED_RHS) == BLOCKED_RHS) {
		self->x = self->oldx;
		self->dx = 0;
		return;
	};
};

void player_death(struct entity_t *self);
	
void player_respawn(entity_t *p) {
	respawn(p);
	
	p->class = CLASS_PLAYER;
	// Set its rendering function appropriately:
	p->next_think = time + 1;
	p->think = (func_t *)player_think;
	p->death = (func_t *)player_death;
	p->render = (func_t *)render_player;
	p->collide = (func_t *)player_collide;

	p->width = 8; p->height = 16;
	// Be invulnerable for 200 frames:
	p->invuln_time = time + 200;

	// Start in the top-left portion of the screen:
	p->x = screen_x + 2048;
	p->y = screen_y + 2048;
};

void msg_render(struct entity_t *self) {
	printf_xy(58, 48, "GAME OVER");
};

void player_death(struct entity_t *self) {
	// Send him upwards and back:
	self->dy = -512;
	self->dx = (self->flags & FLAG_FACERIGHT) ? -1512 : 512;
	// Enable dead and liedown flag, but disable onground flag::
	self->flags |= FLAG_LIEDOWN | FLAG_DEAD;
	self->flags &= ~FLAG_ONGROUND;
	
	if (self->health < 0) {
		self->death = NULL;		// dont call death again!
		self->think = (func_t *)sub_remove;
		self->next_think = time + 90;

		// Create an entity to display a message:
		entity_t *msg = spawn();
		msg->owner = self;
		msg->x = self->x; msg->y = self->y;
		msg->height = 0;
		msg->width = 0;
		msg->flags = FLAG_NOGRAVITY | FLAG_NOFRICTION;
	
		msg->render = (func_t *)msg_render;
		msg->think = (func_t *)sub_remove;
		msg->next_think = time + 90;
	} else {
		// Respawn after 80 frames:
		self->think = (func_t *)player_respawn;
		self->next_think = time + 80;
	};
};

entity_t *player_spawn() {
	entity_t *p = spawn();
	if (p == NULL) return NULL;

	p->class = CLASS_PLAYER;
	// Set its rendering function appropriately:
	p->next_think = 0;
	p->think = (func_t *)player_think;
	p->death = (func_t *)player_death;
	p->render = (func_t *)render_player;
	p->collide = (func_t *)player_collide;

	p->width = 8; p->height = 16;
	p->x = 0;
	p->y = 0;
	
	p->health = p->max_health = 3;
	
	return p;
};
